"""
Message context.

"""

from microcosm.api import defaults, typed
from microcosm.config.types import boolean
from microcosm_logging.decorators import logger

from microcosm_pubsub.errors import TTLExpired


@logger
class SQSMessageContext:
    """
    The message context controls what data you want to associate
    with your daemon handler context, e.g. some combination of keys from the message or
    additional metadata fetched from elsewhere.

    """
    def __init__(self, graph):
        self.enable_ttl = graph.config.sqs_message_context.enable_ttl
        self.initial_ttl = graph.config.sqs_message_context.initial_ttl

    def __call__(self, message_dct, **kwargs):
        context = message_dct.get("opaque_data", dict())

        self.update_context_uri(context, message_dct)
        self.update_context_ttl(context)

        context.update(**kwargs)
        return context

    def update_context_uri(self, context, message_dct):
        # If there is a uri add it to the context
        if message_dct.get("uri"):
            context["uri"] = message_dct.get("uri")

    def update_context_ttl(self, context):
        if self.enable_ttl:
            try:
                ttl = int(context["X-Request-Ttl"])
            except KeyError:
                ttl = self.initial_ttl
            if ttl == 0:
                self.logger.warning(
                    "Error handling SQS message - TTL expired",
                    extra=context,
                )
                raise TTLExpired(extra=context)
            context["X-Request-Ttl"] = str(ttl - 1)


@defaults(
    enable_ttl=typed(boolean, default_value=True),
    initial_ttl=typed(int, default_value=32),
)
def configure_sqs_message_context(graph):
    """
    Configure the message context object.

    Usage:
        graph.message_context()

    """
    return SQSMessageContext(graph)



from inspect import ismethod
from functools import wraps


EXTRACTS = "_extracts"
BINDS = "_binds"


def to_function(callable_object):
    """
    In python, we can't use setattr on a method
    We can instead wrap the method in a function.
    This function will wrap only methods.

    """
    if not ismethod(callable_object):
        return callable_object

    @wraps(callable_object)
    def decorated_method(*args, **kwargs):
        return callable_object(*args, **kwargs)
    return decorated_method


def extracts(*extract):
    """
    On resolving this function with get_from_context decorator -
    updates the context based on the function return value.

    Note: The decorator just marks the function and don't change the return value.
    :param *args: key/s of the return value

    """
    def decorate(func):
        function = to_function(func)
        setattr(function, EXTRACTS, extract)
        return function
    return decorate


def binds(**binds):
    """
    On resolving this function with temporarily_replace_context_keys decorator -
    temporarily updates the context keys.

    Note: The decorator just marks the function.
    :param **binds: (old_context_key, new_context_key) - context keys to rename

    """
    def decorate(func):
        function = to_function(func)
        setattr(function, BINDS, binds)
        return function
    return decorate



"""
PublishBatchMessage handler.

"""

from microcosm.api import binding
from microcosm_logging.decorators import logger
from microcosm_pubsub.conventions import created
from microcosm_pubsub.decorators import handles


@binding("publish_message_batch")
@handles(created("BatchMessage"))
@logger
class PublishBatchMessage:

    def __init__(self, graph):
        self.sns_producer = graph.sns_producer

    def __call__(self, message):
        messages = message["messages"]
        for message in messages:
            self.sns_producer.publish_message(
                message["media_type"],
                message["message"],
                message["topic_arn"],
                message["opaque_data"],
            )



"""
Convention-driven pubsub.

"""
from functools import partial

from microcosm_pubsub.conventions.lifecycle import LifecycleChange
from microcosm_pubsub.conventions.messages import IdentityMessageSchema, URIMessageSchema  # noqa: F401
from microcosm_pubsub.conventions.naming import make_media_type, name_for  # noqa: F401


def media_type(lifecycle_change):
    """
    Fluent wrapper around message publishing for convention-driven schemas.

    """
    return partial(make_media_type, lifecycle_change=lifecycle_change)


def changed(resource, **kwargs):
    return media_type(LifecycleChange.Changed)(resource, **kwargs)


def created(resource, **kwargs):
    return media_type(LifecycleChange.Created)(resource, **kwargs)


def deleted(resource, **kwargs):
    return media_type(LifecycleChange.Deleted)(resource, **kwargs)



from microcosm_pubsub.chain.statements.assign import (  # noqa: F401
    assign,
    assign_constant,
    assign_function,
    extract,
)
from microcosm_pubsub.chain.statements.for_each import for_each  # noqa: F401
from microcosm_pubsub.chain.statements.switch import switch  # noqa: F401
from microcosm_pubsub.chain.statements.try_chain import try_chain  # noqa: F401
from microcosm_pubsub.chain.statements.when import when  # noqa: F401



from functools import wraps, WRAPPER_ASSIGNMENTS
from inspect import signature, _POSITIONAL_ONLY, _POSITIONAL_OR_KEYWORD, _empty

from microcosm_pubsub.chain.decorators import EXTRACTS, BINDS


DEFAULT_ASSIGNED = (EXTRACTS, BINDS)
EXTRACT_PREFIX = "extract_"


def get_positional_args(func):
    """
    Returns all the POSITIONAL argument names of a function
    * If func is a method, do not return "self" argument
    * Handle wrapped functions too

    """
    return [
        (arg, parameter.default)
        for arg, parameter
        in signature(func).parameters.items()
        if parameter.kind in (_POSITIONAL_ONLY, _POSITIONAL_OR_KEYWORD)
    ]


def get_from_context(context, func, assigned=DEFAULT_ASSIGNED):
    """
    Decorate a function - pass to the function the relevant arguments
    from a context (dictionary) - based on the function arg names

    """
    positional_args = get_positional_args(func)

    @wraps(func, assigned=assigned + WRAPPER_ASSIGNMENTS)
    def decorate(*args, **kwargs):
        context_kwargs = {
            arg_name: (context[arg_name] if default is _empty else context.get(arg_name, default))
            for arg_name, default in positional_args[len(args):]
            if arg_name not in kwargs
        }
        return func(*args, **kwargs, **context_kwargs)
    return decorate


def save_to_context(context, func, assigned=DEFAULT_ASSIGNED):
    """
    Decorate a function - save to a context (dictionary) the function return value
    if the function is marked by @extracts decorator

    """
    extracts = getattr(func, EXTRACTS, None)
    if not extracts:
        return func
    extracts_one_value = len(extracts) == 1

    @wraps(func, assigned=assigned + WRAPPER_ASSIGNMENTS)
    def decorate(*args, **kwargs):
        value = func(*args, **kwargs)
        if extracts_one_value:
            value = [value]
        for index, name in enumerate(extracts):
            context[name] = value[index]
        return value
    return decorate


def save_to_context_by_func_name(context, func, assigned=DEFAULT_ASSIGNED):
    """
    Decorate a function - save to a context (dictionary) the function return value
    if the function is not signed by EXTRACTS and it's name starts with "extract_"

    """
    if (
        hasattr(func, EXTRACTS) or
        not hasattr(func, "__name__") or
        not func.__name__.startswith(EXTRACT_PREFIX)
    ):
        return func
    name = func.__name__[len(EXTRACT_PREFIX):]

    @wraps(func, assigned=assigned + WRAPPER_ASSIGNMENTS)
    def decorate(*args, **kwargs):
        value = func(*args, **kwargs)
        context[name] = value
        return value
    return decorate


def temporarily_replace_context_keys(context, func, assigned=DEFAULT_ASSIGNED):
    """
    Decorate a function - temporarily updates the context keys while running the function
    Updates the context if the function is marked by @binds decorator.

    """
    binds = getattr(func, BINDS, None)
    if not binds:
        return func

    @wraps(func, assigned=assigned + WRAPPER_ASSIGNMENTS)
    def decorate(*args, **kwargs):
        for old_key, new_key in binds.items():
            if old_key not in context:
                raise KeyError(f"Variable '{old_key}'' not set")
            if new_key in context:
                raise ValueError(f"Variable '{new_key}'' already set")
        try:
            for old_key, new_key in binds.items():
                context[new_key] = context.pop(old_key)
            return func(*args, **kwargs)
        finally:
            for old_key, new_key in binds.items():
                context[old_key] = context.pop(new_key)
    return decorate
